// Copyright 2015 - Pierre Vuillemin
// License : CECILL

function out = %sopiVar_e(varargin)
   // Overloading of the extraction for a sopiVar.
   //
   // Inputs
   // ------
   // varargin(1) (int or matrix or string): the row elements or field to extract
   // [varargin(2) (int or matrix): the column elements to extract]
   // varargin($) (sopiVar): the sopiVar
   //
   //
   // Returns
   // -------
   // The desired field or answer to the test.
   //
   // Notes
   // -----
   // The sopiVar is the last element of the input list, i.e. varargin($). Then
   //
   // - if there are only 2 inputs and the first one is a string, one needs to
   // extract some field from the sopiVar and sopi_extractField is called,
   //
   // - otherwise (and if there are less than 3 inputs, as sopiVar can only be
   // matrices) one needs to extract some elements from the sopiVar and 
   // sopi_extractElement is called.
   var = varargin($)
   if length(varargin) == 2 & typeof(varargin(1)) == "string" then
      out = sopi_extractField(var,varargin(1))
   elseif length(varargin) <= 3 then
      out = sopi_extractElement(var,varargin(1:$-1));
   else
      error("Invalid extraction")
   end
endfunction

function out = sopi_extractField(var, field)
   // Extraction of a field from a sopiVar
   //
   // Inputs
   // ------
   // var (sopiVar): the sopiVar
   // field (string): the name of the field to extract
   //
   // Returns
   // -------
   // The desired field or answer to the test.
   //
   // Notes
   // -----
   // If the desired field is an actual field of the sopiVar, then it is returned.
   // Otherwise, an some tests are performed on the sopiVar through sopi_performTest. 
   // These tests could be seen as artificial fields, or shortcuts.
   validFieldNames = fieldnames(var)
   if or(field == validFieldNames) then
      out = getfield(field, var)
   else
      out = sopi_performTest(var, field)
   end
endfunction

function out = sopi_performTest(var, test)
   // Performs some tests on a sopiVar.
   //
   // Inputs
   // ------
   // var (sopiVar): the sopiVar
   // test (string): the name of the test to perform
   //
   // Returns
   // -------
   // The answer of the test.
   //
   // Notes
   // -----
   // If none of the direct tests matches, then one attempts to perform more
   // advanced tests on the mathematical properties of the sopiVar by parsing
   // it and by testing the result (see %sopiProp_e).
   //
   select test
   case "isElementary"
      out = isempty(var.operator)
   case "isSecondary"
      out = ~var.isElementary
   case "isScalar"
      out = isscalar(var)
   case "isSquare"
      out = size(var,1) == size(var,2)
   case "depends"
      out = sopi_getVarDependency(var)
   else
       try
           [m,n]    = size(var)
           out      = zeros(m,n) == 0
           for i = 1:m
               for j = 1:n
                   out(i,j) = var.prop(i)(j)(test)
               end
           end
       catch
           error("Invalid field!" + test)
       end
   end
endfunction

function newVar = sopi_extractElement(varargin)
   // Extracts some elements (in a matrix sense) of a sopiVar.
   //
   // Inputs
   // ------
   // varargin(1) (int or matrix): the rows to extract
   // [varargin(2) (int or matrix): the columns to extract]
   // varargin($) (sopiVar): the sopiVar
   //
   // Returns
   // -------
   // A sopiVar containing a tree representing the operation.
   //   
   var      = varargin(1);
   [m, n]   = size(var);
   // row extraction
   i        = varargin(2);
   Im       = eye(m,m);
   Ei       = Im(i,:)
   if Ei == Im then
      if length(varargin) == 2 
         // vectorisation
         newVar = sopi_vec(var);
      else
         // all the rows are required, i.e. x(:,columnID)
         newVar = var;
      end
   else
      // only some rows are required
      newVar = Ei * var;
   end
   if length(varargin) == 3 then
      // column extraction
      j  = varargin(3);
      In = eye(n,n);
      Ej = In(:,j);
      if Ej ~= In then
         // only some columns are required
         newVar  = newVar * Ej;
      end
   end
endfunction

function newVar = sopi_vec(var)
   // Vectorize a sopiVar.
   //
   // Inputs
   // ------
   // var(sopiVar): the sopiVar
   //
   // Returns
   // -------
   // A sopiVar containing a tree representing the operation.
   //      
   if var.isScalar |  size(var,2) == 1 then
      newVar = var;
   elseif size(var,1) == 1 & ~var.isScalar then
      newVar = var';
   else
      newVar            = sopi_newVar()
      newVar.size       = [size(var,1)*size(var,2),1]
      newVar.operator   = "vec"
      newVar.args       = list(var)      
   end
endfunction
