// Author : Pierre Vuillemin (2016)
// License : CECILL

function out = sopi_extractLinearPart(var)
   // Extracts the linear elemens of a variable.
   //
   // Input
   // -----
   // var (sopi_Var): the variable to analyse.
   // 
   // Returns
   // -------
   // A list containing the left and right factors together with the elementary
   // variable(s) on which it is applied.
   //
   // Example
   // -------
   // sopi_begin()
   // x        = sopi_var(3,3);
   // x.name   = "x"
   // A        = rand(3,3);
   // B        = rand(3,3);
   // C        = rand(3,3);
   // z        = A * x * B + C;
   // out      = sopi_extractLinearPart(z);
   // out(1)(1) == A
   // out(1)(2) == B
   // out(1)(3)
   //
   if var.isElementary then
      // if the variable is elementary then its left and right linear parts are 
      // identity matrices, i.e. z = I_1 * z * I_2
      [m, n] = size(var)
      L      = eye(m,m)
      R      = eye(n,n)
      out    = list(list(L, R, var))
   else
      select var.operator
      case "times"
         // When z = x * y
         if var.args(1).isConstant then
            // If x is a constant then one seeks the linear parts in y. 
            // The linear parts of z are then given as:
            // - the linear parts of y on the right
            // - x * the left linear parts of y on the left.
            L     = var.args(1).args(1)
            inner = sopi_extractLinearPart(var.args(2))
            out   = list()
            for i = 1:size(inner)
               out(i) = list(L * inner(i)(1), inner(i)(2), inner(i)(3))
            end
         elseif var.args(2).isConstant then
            // Similarly on the other side.
            R     = var.args(2).args(1)
            inner = sopi_extractLinearPart(var.args(1))
            out   = list()
            for i = 1:size(inner)
               out(i) = list(inner(i)(1), inner(i)(2) * R, inner(i)(3))
            end
         else
            // If neither x nor y are constants then
            // - the left parts of z are made by the constant part of x times the
            // left linear parts of y
            // - the right linear parts are made by the right linear parts of x 
            // times the constant part of y
            L0    = sopi_extractConstantPart(var.args(1))
            R0    = sopi_extractConstantPart(var.args(2))
            L1    = sopi_extractLinearPart(var.args(1))
            R1    = sopi_extractLinearPart(var.args(2))
            for i = 1:size(L1)
               L1(i)(2) = L1(i)(2) * R0
            end
            for i = 1:size(R1)
               R1(i)(1) = L0 * R1(i)(1)
               out = lstcat(L1,R1)
            end
         end
      case "sum"
         // if z = x + y, the linear parts of x and y are appended
         out = list()
         for i = 1:length(var.args)
            inner = sopi_extractLinearPart(var.args(i))
            if ~isempty(inner)
               out   = lstcat(out, inner)
            end
         end
      case "constant"
         // if z = constant, then it has no linear part
         out = []
      else
         disp(sprintf("Cannot extract linear part from: ""%s"" ",var.operator))
         out = []
         return
      end
   end
   if length(out) > 1 then
      // when multiple linear depedencies are extracted from a variable, some of 
      // them might be redundant, for instance when z = A*x + B*x.
      // Here the redundant variable depedencies are 'factorized'.
      i = 1
      next = %t
      while next
         varId = out(i)(3).id
         for j = size(out):-1:i+1
            if out(j)(3).id == varId then
               left_equal  = and(out(i)(1) == out(j)(1))
               right_equal = and(out(i)(2) == out(j)(2))
               if left_equal & right_equal then
                  // A  x  B + A  x  B = 2 * A  x  B
                  out(i)(1) = 2 * out(i)(1)
               elseif left_equal & ~right_equal then
                  // A x B + A x C = A x (B + C)
                  out(i)(1) = out(i)(1)
                  out(i)(2) = out(i)(2) + out(j)(2)
               elseif ~left_equal & right_equal then
                  // A x B + C x B = (A + C) x B
                  out(i)(1) = out(i)(1) + out(j)(1)
                  out(i)(2) = out(i)(2)
               else
                  // A x B + C x D = (A + C) x (B + D)
                  out(i)(1)   = out(i)(1) + out(j)(1)
                  out(i)(2)   = out(i)(2) + out(j)(2)
               end
               out(j)      = null()
            end
         end
         if i == size(out) then
            next = %f
         end
         i = i+1
      end
   end
endfunction

