// Author : Pierre Vuillemin (2016)
// License : CECILL

function problem = sopi_min(fun, varargin)
   // Creates an optimisation problem.
   //
   // Calling Sequence
   // problem = sopi_min(fun)
   // problem = sopi_min(fun, c1, ..., cn)
   //
   // Parameters
   // fun: a sopiVar representing the objective function
   // ci: constraints given as sopiCst or list of sopiCst
   // problem: a sopiPro representing the optimisation problem
   //
   // See also
   // sopi_solve()
   // sopi_example()
   if ~isscalar(fun) then
      error("Multi-objective optimisation is not handled")
   end
   cList = list()
   for c = varargin
      if typeof(c) == "sopiCst" then
         cList($+1) = c;
      elseif typeof(c) == "list" then
         cList = lstcat(cList, c)
      else
         error("Unexpected input")
      end
   end
   //
   problem = sopi_newProblem()
   problem = sopi_addAllVar(problem, fun, cList)
   problem = sopi_addObjective(problem, fun)
   problem = sopi_addConstraints(problem, cList)
   // Optimisation problem classification
   problem.class(1) = problem.objType
   if isempty(cList) then
      // The optimisation problem is unconstrained
      problem.class(2) = -1
   else
      // The optimisation problem is constrained
      if and(problem.cType == 0)
         // There are only bounds constraints
         problem.class(2) = 0
      elseif and(problem.cType <= 1)
         // There are only linear constraints
         problem.class(2) = 1
      end
   end
   problem.solvers = sopi_getSolver(problem.class)
endfunction

function out = sopi_getSolver(class)
   out      = list()
   solvers  = sopi_listSolvers()
   for s = solvers
      if or(s.objClass == problem.class(1)) & or(s.cstClass == problem.class(2))
         out($+1) = s
      end
   end
endfunction
