// Author : Pierre Vuillemin (2016)
// License : CECILL

function [outVar, fopt, info] = sopi_solve(problem, varargin)
   // Solves an optimisation problem.
   //
   // Calling Sequence
   // sopi_solve(problem, method)
   //
   // Parameters
   //    problem; a sopiPro
   //    method: the name of the optimisation method to use
   //
   // Examples
   //  
   // // -----------------------------------------------------------------------------
   // //                                  SOPI DEMO
   // // -----------------------------------------------------------------------------
   // // This example shows how Sopi can be used to formulate and solve a simple linear
   // // optimisation problem.
   // //
   // // The considered problem is the following,
   // //       min_x,y        10 x + 12 y
   // //       s.t.
   // //                      y <= 2 x
   // //                2x + 3y <= 240
   // //                  x + y >= 60
   // //                   x, y >= 0 
   // //
   // // The optimal solution is reached at x = 60 and y = 0 for f(x,y) = 600.
   // // -----------------------------------------------------------------------------
   //
   // // sopi_begin must be called before any sopi-related operation
   // sopi_begin()
   // // Creation of the optimisation variables x and y
   // x        = sopi_var(1)
   // y        = sopi_var(1)
   // x.name   = "x"
   // y.name   = "y"
   // // Creation of the constraints of the problem. Note that constraints can be 
   // // gathered with &
   // bounds   = x >= 0 & y >=0
   // c1       = y <= 2 * x
   // c2       = 2*x +3*y <= 240
   // c3       = x+y >= 60
   // // Creation of the objective function
   // objective = 10*x + 12*y
   // // Instanciation of the optimisation problem
   // problem           = sopi_min(objective, bounds, c1, c2, c3)
   // disp(sprintf("Optimisation problem built by sopi:\n"))
   // disp(problem)
   // // Resolution of the problem
   // // At the moment, the default linear solver is the built-in method 'karmarkar'.
   // [optVar, fopt, info] = sopi_solve(problem)
   // // The optimal values are given in the fields 'opt'
   // disp(sprintf("Solution of the problem:\n"))
   // disp(sprintf('\tSolver used: ''%s''', problem.solvers(1).name))
   // disp(sprintf('\tOptimal value of %s:%2.2f', x.name, x.opt))
   // disp(sprintf('\tOptimal value of %s:%2.2f', y.name, y.opt))
   // disp(sprintf('\tOptimal objective value:%2.2f', objective.opt))
   // // Displaying the problem and the solution found
   // sopi_plot2DProblem(problem, x, y)
   // sopi_end() 
   tic()
   // Optional arguments handling
   method   = []
   x0       = []
   options  = []
   for i = 1:length(varargin)
      select typeof(varargin(i))
      case 'string'
         method = varargin(i)
      case 'constant'
         x0 = varargin(i)
      case 'st'
         options = varargin(i)
      end
   end
   // Solver
   if isempty(method) then
      solver = problem.solvers(problem.defaultSolver)
   else
      solverAvailable = %F
      for s = problem.solvers
         if s.name == method
            solverAvailable   = %T
            solver            = s
            break;
         end
      end
      if ~solverAvailable
         error('The method ''%s'' is not available problem of the type ''%s''',method,sopi_verboseProblemClass(problem.class))
      end
   end
   // Testing if the solver require x0
   if isempty(x0) & solver.requireX0 then
      sopi_print(0,"The solver ''%s'' requires an initial point. Choosing one for you.\n",solver.name)
      x0 = ones(problem.nvar,1)
   end
   //
   if isempty(options) then
      options = struct()
   end
   // Calling the solver
   execstr("sol = " + solver.call + "(problem, x0, options)")
   // Handling the outputs
   optVar = list()
   for v = problem.varList
      idx         = sopi_getVarIdxInProblem(problem, v)
      v.opt       = sol.xopt(idx)
      optVar($+1) = v
   end
   outVar            = optVar
   fopt              = sol.fopt
   info              = struct()
   info.elapsedTime  = toc();
   // Assigning output optimal values
   clear v
   objective      = problem.objective
   objective.opt  = sol.fopt
   optVar($+1)    = objective
   exVars   = sopi_getExistingSopiVar()
   couples  = list()
   for v = exVars
      for vopt = optVar
         if vopt.id == v(1).id
            couples($+1) = list(v(2), vopt)
         end
      end
   end
   out      = [];
   arg      = [];
   for i = 1:length(couples)
      out = [out,couples(i)(1)]
      arg = [arg,"couples(" + string(i) + ")(2)"];
   end
   exStr = "[" + strcat(out,',') + "] = resume(" + strcat(arg,",")+");";
   execstr(exStr);
endfunction
